/* Solution in C# by Jimmy Mrdell */

using System;
using System.Collections;
using System.IO;

namespace LazeEvaluation
{
	// Keeps statistics about an evaluation
	public class EvalStat : ICloneable
	{
		public int nAdd = 0, nSub = 0, nMult = 0, nDiv = 0, nRem = 0, nFuncEval = 0;
		public EvalStat() { }
		public object Clone() { return this.MemberwiseClone(); }
	}
	
	// Exception thrown when caught in an infinite loop
	public class InfiniteLoopException : Exception { }

	// Abstract base class for expressions.
	// Inherited by FuncExpression and ConstantExpression
	public abstract class Expression
	{
		public abstract ConstantExpression Evaluate(bool lazy, EvalStat evalStat);
		
		// Parse the input data
		public static Expression Parse(ref string line)
		{
			if (line.Length==0) throw new Exception("Bad expression");
			if (line[0]=='(')
			{
				int i = 1, cnt = 1;
				while (i<line.Length && cnt>0)
				{
					if (line[i]=='(') cnt++;
					if (line[i]==')') cnt--;
					i++;
				}
				if (cnt>0) throw new Exception("Missing final )");
				string param = line.Substring(1,i-2);
				line = line.Remove(0,i);
				Expression funcName = Expression.Parse(ref param);
				ArrayList paramList = new ArrayList();
				while (param.Length>0)
				{
					param = param.Substring(1); // Remove separating space
					paramList.Add(Expression.Parse(ref param));
				}
				return new FuncExpression(funcName,(Expression[])paramList.ToArray(typeof(Expression)));
			}
			int delim = line.IndexOf(' ');
			if (delim<0)
				delim = line.Length;
			string s = line.Substring(0,delim);
			line = line.Substring(delim);
			if (delim==0) throw new Exception("Bad expression");
			if (char.IsLetter(s[0]))
				return new ConstantExpression(s);
			return new ConstantExpression(int.Parse(s));
		}
	}

	// Wrapper expression class for constant (including names)
	public class ConstantExpression : Expression
	{
		public object Value;
		public ConstantExpression(object value) { this.Value = value; }
		public override ConstantExpression Evaluate(bool lazy, EvalStat evalStat) { return this; }
		public override string ToString() { return Value.ToString(); }
	}

	public class FuncExpression : Expression
	{
		public Expression Name;
		public Expression[] Params;
		private ConstantExpression _evaluatedValue;

		// For debugging
		public override string ToString()
		{
			string s = "(" + Name.ToString();
			foreach(Expression e in Params)
				s = s + " " + e.ToString();
			return s + ")";
		}
		
		public FuncExpression(Expression name, Expression[] param)
		{
			this.Name = name;
			this.Params = param;
			_evaluatedValue = null;
		}

		public override ConstantExpression Evaluate(bool lazy, EvalStat evalStat)
		{
			//string evalString = this.ToString();
			//Console.WriteLine("Evaluating: " + evalString);

			if (_evaluatedValue==null)
			{
				if (evalStat.nFuncEval++>10000)
					throw new InfiniteLoopException();

				string name = Name.Evaluate(lazy, evalStat).ToString();

				if (!lazy) // Evaluate parameters if strict evaluation
					for(int i=0; i<Params.Length; i++)
						Params[i] = Params[i].Evaluate(lazy, evalStat);

				// Check short-circuited functions
				if (name=="true")
					_evaluatedValue = Params[0].Evaluate(lazy, evalStat);
				else if (name=="false")
					_evaluatedValue = Params[1].Evaluate(lazy, evalStat);
				else
				{
					// Check user-defined functions
					Function f = Function.FunctionCache[name] as Function;
					if (f!=null)
					{
						Expression e = f.ApplyParams(Params);
						//Console.WriteLine("Expanded to: " + e);
						_evaluatedValue = e.Evaluate(lazy, evalStat);
					}
					else
					{
						object eval;
						int v1 = (int)Params[0].Evaluate(lazy, evalStat).Value;
						int v2 = (int)Params[1].Evaluate(lazy, evalStat).Value;

						switch (name)
						{
							case "add" : eval = v1 + v2; evalStat.nAdd++; break;
							case "sub" : eval = v1 - v2; evalStat.nSub++; break;
							case "mult": eval = v1 * v2; evalStat.nMult++; break;
							case "div" : eval = v1 / v2; evalStat.nDiv++; break;
							case "rem" : eval = v1 % v2; evalStat.nRem++; break;
							case "eq"  : eval = v1 == v2 ? "true" : "false"; break;
							case "gt"  : eval = v1 > v2 ? "true" : "false"; break;
							default : throw new Exception("Unknown function: " + name);
						}
						_evaluatedValue = new ConstantExpression(eval);
					}
				}
			}

			//Console.WriteLine("Evaluation done: " + evalString + " = " + _evaluatedValue);

			return _evaluatedValue;
		}
	}

	// Class for user defined function
	public class Function
	{
		// A hash table over all user defined functions
		public static Hashtable FunctionCache = new Hashtable();

		private string _name;
		private string[] _formalParams;
		private Expression _definition;

		public Function(string name, string[] formalParams, string expression)
		{
			this._name = name;
			this._formalParams = formalParams;
			this._definition = Expression.Parse(ref expression);
		}

		public Expression ApplyParams(Expression[] actualParams)
		{
			return ApplyParams(_definition, actualParams);
		}

		private Expression ApplyParams(Expression expr, Expression[] actualParams)
		{
			if (actualParams.Length!=_formalParams.Length)
				throw new Exception("Mismatch in number of parameters in " + this._name);

			if (expr is FuncExpression)
			{
				FuncExpression fe = expr as FuncExpression;	
				Expression[] appliedParams = new Expression[fe.Params.Length];
				for(int i=0; i<fe.Params.Length; i++)
					appliedParams[i] = ApplyParams(fe.Params[i], actualParams);
				return new FuncExpression(ApplyParams(fe.Name, actualParams), appliedParams);
			}

			string constant = (expr as ConstantExpression).Value as string;
			for(int i=0; i<_formalParams.Length; i++)
				if (constant == _formalParams[i]) return actualParams[i];
			return expr;
		}
	}
	
	// MAIN
	class LazyEvaluationMain
	{
		[STAThread]
		static void Main(string[] args)
		{
			//TextReader input = new StreamReader(@"d:\code\acm\2004\lazy\lazy-test.in");
			TextReader input = Console.In;
			
			string expr;
			while ((expr = input.ReadLine()) != "")
			{
				int ix1 = expr.IndexOf('='), ix2 = expr.IndexOf(' ');
				string[] formalParams = ix1-ix2<2 ?
					new string[] { } : expr.Substring(ix2+1,ix1-ix2-2).Split(new char[] { ' ' } );
				string name = expr.Substring(0,ix2);
				Function func = new Function(name, formalParams, expr.Substring(ix1+2));
				Function.FunctionCache.Add(name, func);
			}

			EvalStat lazyStat = new EvalStat(), strictStat = new EvalStat();
			while (true)
			{
				expr = input.ReadLine();
				if (expr==null || expr=="") break;

				EvalStat bakLazyStat = lazyStat.Clone() as EvalStat;
				EvalStat bakStrictStat = strictStat.Clone() as EvalStat;
				
				try
				{
					// Lazy evaluation
					lazyStat.nFuncEval = 0; // Reset evaluation counter
					string lazyExprString = expr; // Create shallow copy
					Expression lazyExpr = Expression.Parse(ref lazyExprString);
					string lazyAns = lazyExpr.Evaluate(true, lazyStat).ToString();
				
					// Strict evaluation
					strictStat.nFuncEval = 0;
					string strictExprString = expr;
					Expression strictExpr = Expression.Parse(ref strictExprString);
					string strictAns = strictExpr.Evaluate(false, strictStat).ToString();
				}
				catch (InfiniteLoopException)
				{
					lazyStat = bakLazyStat;
					strictStat = bakStrictStat;
				}
			}

			Console.WriteLine("operator lazy_evaluation strict_evaluation");
			Console.WriteLine("add    {0,8}         {1,8}", lazyStat.nAdd,  strictStat.nAdd);
			Console.WriteLine("sub    {0,8}         {1,8}", lazyStat.nSub,  strictStat.nSub);
			Console.WriteLine("mult   {0,8}         {1,8}", lazyStat.nMult, strictStat.nMult);
			Console.WriteLine("div    {0,8}         {1,8}", lazyStat.nDiv,  strictStat.nDiv);
			Console.WriteLine("rem    {0,8}         {1,8}", lazyStat.nRem,  strictStat.nRem);
		}
	}
}
